/*
 * Copyright (c) 1995-1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/x86/pc/rtc.h>		/* RTC_BASEHI */
#include <oskit/x86/pc/phys_lmm.h>	/* phys_lmm_add */
#include <oskit/x86/types.h>		/* oskit_addr_t */
#include <oskit/x86/base_cpu.h>		/* base_cpu_setup */
#include <oskit/x86/base_vm.h>		/* kvtophys */

#include <oskit/c/ctype.h>		/* isspace */
#include <oskit/c/stdlib.h>		/* malloc */
#include <oskit/c/string.h>		/* strcpy */
#include <oskit/c/stdio.h>		/* printf */
#include <oskit/c/assert.h>		/* assert */
#include <oskit/c/arpa/inet.h>		/* ntohl */
#include <oskit/c/malloc.h>		/* malloc_lmm */
#include <oskit/lmm.h>			/* lmm_remove_free */
#include <oskit/x86/pc/base_multiboot.h>	/* boot_info */

#include <oskit/dev/dev.h>

#include "getline.h"			/* getline */
#include "timer.h"			/* TIMER_FREQ */
#include "driver.h"			/* net_init */
#include "ether.h"			/* arptable */
#include "gdt.h"			/* copy_base_gdt */
#include "boot.h"			/* kerninfo */
#include "netboot.h"			/* netprintf */
#include "reboot.h"			/* reboot_stub_loc */

extern oskit_addr_t return_address;
extern char version[], build_info[];

static void
init_default_args(int argc, char *argv[], char *default_args[])
{
	char **p;
	int i;

	/*
	 * Figure out if -h (serial) or -f (fast serial) were given
	 * and put them in the default_args list.
	 */
	p = &default_args[0];
	for (i = 0; i < argc; i++)
#define NDEFAULT_ARGS 2
		if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "-f") == 0)
			*p++ = argv[i];
	*p = NULL;
}

static void
massage_args(char **argstringp, char *default_args[NDEFAULT_ARGS])
{
	int i;
	char *p;

	/*
	 * If no argstring, build one out of the default_args.
	 */
	if (*argstringp == NULL) {
		int size;

		size = 0;
		for (i = 0; i < NDEFAULT_ARGS && default_args[i]; i++)
			size += strlen(default_args[i]) + 1;	/* +1 for ' ' */

		*argstringp = mustcalloc(size + 1, 1);		/* +1 for NUL */
		goto copyem;
	}

	/*
	 * Look for ones that they specified to toggle off.
	 * If they specified one of the default_args,
	 * we turn it off by  "erasing" it from the argstring and
	 * the default_args.
	 */
	for (i = 0; i < NDEFAULT_ARGS && default_args[i]; i++) {
		p = strstr(*argstringp, default_args[i]);
		if (p) {
			/* "Erase" it. */
			memset(p, ' ', strlen(default_args[i]));
			default_args[i] = "";
		}
	}

	/*
	 * Copy the remaining default_args into the argstring.
	 */
 copyem:
	p = mustmalloc(strlen(*argstringp)
		       + strlen(" -retaddr 0x00ff00ff") + 1);
	sprintf(p, "%s -retaddr 0x%08x", *argstringp, reboot_stub_loc);
	*argstringp = p;
	for (i = 0; i < NDEFAULT_ARGS && default_args[i]; i++) {
		strcat(*argstringp, " ");
		strcat(*argstringp, default_args[i]);
	}
}

void
main(int argc, char *argv[])
{
	char buf[512];
	oskit_error_t err;
	int done = 0;
	struct kerninfo kinfo;
	struct multiboot_info *mbi;
	char *cmdline = buf;
	int i;
	char *default_args[NDEFAULT_ARGS] = {0};
	char *argstring;

	printf("%s\n", version);

	oskit_dev_init();
	osenv_timer_register(bumpticks, TIMER_FREQ);

	err = net_init();
	if (err)
		panic("Couldn't initialize network: 0x%08x", err);
	netprintf("I am %s (IP: %I, GW: %I, netmask: %I)\r\n",
		  hostname, arptable[ARP_CLIENT].ipaddr,
		  arptable[ARP_GATEWAY].ipaddr, netmask);

	init_default_args(argc, argv, default_args);

	/*
	 * Figure out our rootserver, dir to mount, and file to load.
	 * Then try to get it.
	 */
	printf("\nType \"help\" for help.\n\n");
reprompt:
	printf("NetBoot> ");
	getline(cmdline, 512);
	cmdline[strlen(cmdline) - 1] = '\0';	/* chop \n */
	while (cmdline[0] && isspace(cmdline[0]))
		cmdline++;
	if (strcmp(cmdline, "help") == 0) {
		printf("\n");
		printf("%s\n", version);
		printf("%s\n", build_info);
		printf("\n");
		printf("To boot, tell me the kernel file name and args in the form\n"
		       "\t<IPaddr|host>:<path> [args]\n"
		       "For example:\n"
		       "\t155.22.33.44:/home/foo/kernel -f -h -wow\n"
		       "or\n"
		       "\tmyhost:/home/foo/kernel -f -h -yay\n"
		       "The directory must be an NFS exported directory.\n"
		       "The kernel must be a MultiBoot kernel.\n"
		       "Hostnames are currently limited to \"marker\" and \"fast\".\n");
		printf("\n");
		printf("Type \"quit\" or \"exit\" to %s.\n",
		       return_address? "return to caller" : "reboot");
		if (default_args[0]) {
			printf("\n");
			printf("The default args are as follows and act as toggles:\n");
			printf("\t");
			for (i = 0; i < NDEFAULT_ARGS && default_args[i]; i++)
				printf("%s ", default_args[i]);
			printf("\n");
		}
		printf("\n");
		goto reprompt;
	}
	if (strcmp(cmdline, "quit") == 0 || strcmp(cmdline, "exit") == 0) {
		done = 1;
		goto done;
	}
	if (cmdline[0] == '\0')
		goto reprompt;
	if (! parse_cmdline(cmdline,
			    &kinfo.ki_ip.s_addr,
			    &kinfo.ki_dirname,
			    &kinfo.ki_filename,
			    &argstring)) {
		printf("Bad command line, try again.\n");
		goto reprompt;
	}
	massage_args(&argstring, default_args);
	netprintf("Root server: %I, dir: %s,\r\n"
		  "  file: %s,  cmdline: `%s'\r\n",
		  kinfo.ki_ip.s_addr, kinfo.ki_dirname, kinfo.ki_filename,
		  argstring ? argstring : "");
	memcpy(&arptable[ARP_ROOTSERVER].ipaddr, &kinfo.ki_ip.s_addr,
	       sizeof arptable[ARP_ROOTSERVER].ipaddr);
	if (getkernel_net(&kinfo) != 0) {
		printf("Can't get the kernel, try again.\n");
		goto reprompt;
	}


	/*
	 * Clean some stuff up before we load the new image.
	 * In particular we don't want the timer hammering it,
	 * and don't want it using some bogus GDT (we may overwrite ours
	 * with the new image).
	 */
 done:
	net_shutdown();
	osenv_timer_shutdown();
	copy_base_gdt();

	if (done)
		exit(0);

	/*
	 * Reserve the memory where the kernel will eventually reside.
	 * Future malloc's, like the one for the multiboot_info struct
	 * we pass to the new kernel, will stay out of this region.
	 *
	 * IMPORTANT NOTE: We must do this AFTER the shutdown stuff
	 * because all lmm_remove_free does is alloc all the free
	 * blocks in the given range.
	 * If things were already allocated in that range but get freed
	 * after the lmm_remove_free, they can be allocated again and
	 * then get clobbered by the kernel copy in do_boot and Bad Things
	 * will happen.
	 */
	lmm_remove_free(&malloc_lmm,
			(void *)phystokv(kinfo.ki_mbhdr.load_addr),
			kerninfo_imagesize(&kinfo));

	/*
	 * Allocate and fill in a multiboot_info struct we will pass.
	 */
	mbi = kinfo.ki_mbinfo = mustcalloc(sizeof(struct multiboot_info), 1);
	mbi->mem_lower = boot_info.mem_lower;
	mbi->mem_upper = boot_info.mem_upper;
	mbi->flags |= MULTIBOOT_MEMORY;
	if (argstring) {
		mbi->cmdline = kvtophys(strdup(argstring));
		mbi->flags |= MULTIBOOT_CMDLINE;
	}
	
	/* Load it. */
	loadkernel(&kinfo);
	panic("loadkernel returned");
}
