/*
 * Copyright (c) 1997-1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */
/*
 * This file implements some of the methods for "console" fd's.
 * All this is is a big hack so that 'write(1, ' actually shows up
 * on the console.
 */

#include <stdio.h>
#include <errno.h>
#include <fd.h>
#include <oskit/io/asyncio.h>
#include <oskit/machine/pc/base_console.h>

static struct oskit_stream stream;
static struct oskit_asyncio asyncio;

static OSKIT_COMDECL
query(oskit_stream_t *si, const struct oskit_guid *iid, void **out_ihandle)
{
	if (memcmp(iid, &oskit_iunknown_iid, sizeof(*iid)) == 0 ||
	    memcmp(iid, &oskit_stream_iid, sizeof(*iid)) == 0) {
		*out_ihandle = &stream;
		return 0;
	}

	if (memcmp(iid, &oskit_asyncio_iid, sizeof(*iid)) == 0) {
		*out_ihandle = &asyncio;
		return 0;
	}

	*out_ihandle = NULL;
	return OSKIT_E_NOINTERFACE;
}

static OSKIT_COMDECL_U
addref(oskit_stream_t *si)
{
	return 1;
}

static OSKIT_COMDECL_U
release(oskit_stream_t *si)
{
	return 1;
}

static OSKIT_COMDECL
read(oskit_stream_t *si, void *buf, oskit_u32_t len, oskit_u32_t *out_actual)
{
	char *cbuf = buf;
	int actual = 0;
	int ch;

	if (len == 0)
		return 0;

	/* implement simple line discipline: return newline */
	while (len > 0 && (ch = console_getchar()) >= 0) {
		*cbuf++ = ch;
		actual++;
		len--;
		if (ch == '\n')
			break;
	}
	if (actual > 0) {
		*out_actual = actual;
		return 0;
	}
	return errno;
}

static OSKIT_COMDECL
write(oskit_stream_t *si, const void *buf, oskit_u32_t nb,
      oskit_u32_t *out_actual)
{
	console_putbytes((const char *) buf, (int) nb);

	*out_actual = nb;
	return 0;
}

static OSKIT_COMDECL
seek(oskit_stream_t *si, oskit_s64_t ofs, oskit_seek_t whence,
		  oskit_u64_t *out_newpos)
{
	return OSKIT_ESPIPE;
}

static OSKIT_COMDECL
setsize(oskit_stream_t *si, oskit_u64_t new_size)
{
	return OSKIT_EINVAL;
}

static OSKIT_COMDECL
copyto(oskit_stream_t *s, oskit_stream_t *dst,
       oskit_u64_t size,
       oskit_u64_t *out_read,
       oskit_u64_t *out_written)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
commit(oskit_stream_t *si, oskit_u32_t commit_flags)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
revert(oskit_stream_t *si)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
lockregion(oskit_stream_t *si, oskit_u64_t offset,
			 oskit_u64_t size, oskit_u32_t lock_type)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
unlockregion(oskit_stream_t *si, oskit_u64_t offset,
			   oskit_u64_t size, oskit_u32_t lock_type)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
stat(oskit_stream_t *si, oskit_stream_stat_t *out_stat,
		  oskit_u32_t stat_flags)
{
	return OSKIT_E_NOTIMPL;
}

static OSKIT_COMDECL
clone(oskit_stream_t *si, oskit_stream_t **out_stream)
{
	return OSKIT_E_NOTIMPL;
}

static struct oskit_stream_ops stream_ops = {
	query, addref, release,
	read, write, seek, setsize, copyto,
	commit, revert, lockregion, unlockregion, stat, clone
};

/*
 * this should trigger applications who select on a console fd to
 * go into read->getchar or write->putchar immediately.
 */
static OSKIT_COMDECL
asyncio_poll(oskit_asyncio_t *f)
{
	return OSKIT_ASYNCIO_WRITABLE | OSKIT_ASYNCIO_READABLE;
}

/* add a listener */
static OSKIT_COMDECL
asyncio_add_listener(oskit_asyncio_t *f, struct oskit_listener *l,
	oskit_s32_t mask)
{
	return OSKIT_E_NOTIMPL;
}

/* remove a listener */
static OSKIT_COMDECL
asyncio_remove_listener(oskit_asyncio_t *f, struct oskit_listener *l0)
{
	return OSKIT_E_NOTIMPL;
}

/* how many bytes can be read */
static OSKIT_COMDECL
asyncio_readable(oskit_asyncio_t *f)
{
	return 0;
}

static struct oskit_asyncio_ops asyncio_ops = {
	query, addref, release,	/* will generate three leg. warnings */
	asyncio_poll,
	asyncio_add_listener,
	asyncio_remove_listener,
	asyncio_readable
};

oskit_error_t
fd_set_console(void)
{
	stream.ops = &stream_ops;
	asyncio.ops = &asyncio_ops;

	fd_lock();
	fd_free(0); fd_alloc((oskit_iunknown_t*)&stream, 0);
	fd_free(1); fd_alloc((oskit_iunknown_t*)&stream, 1);
	fd_free(2); fd_alloc((oskit_iunknown_t*)&stream, 2);
	fd_unlock();

	return 0;
}

