/*
 * Copyright (c) 1997-1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */
/*
 * This header file defines common types and functions used by
 * the minimal C library's default POSIX file descriptor layer implementation
 * (read, write, seek, connect, accept, etc.).
 * This header file is NOT a standard POSIX or Unix header file;
 * instead its purpose is to expose the implementation of this facility
 * so that the client can fully control it and use it in arbitrary contexts.
 * 
 * Since the minimal C library does not know the context it is used in,
 * you must initialize it with references to appropriate COM objects
 * which provide the actual underlying services.
 */
#ifndef _OSKIT_C_FD_H_
#define _OSKIT_C_FD_H_

#include <oskit/compiler.h>
#include <oskit/c/stddef.h>
#include <oskit/net/socket.h>
#include <oskit/fs/dir.h>
#include <oskit/fs/openfile.h>
#include <oskit/c/sys/time.h>

/*
 * Structure representing a file descriptor table entry.
 * A file descriptor is basically just a generic COM object reference;
 * for efficiency it can also hold cached refs to specific interfaces.
 */
struct fd {
	/*
	 * This is the primary object reference in the file descriptor;
	 * it is non-null if the fd is in use, or null if not.
	 */
	struct oskit_iunknown	*obj;

	/*
	 * Each of these can only be non-null if obj is non-null;
	 * if they are non-null, they are simply cached references
	 * to particular interfaces previously queried for on 'obj'.
	 * If any new cached interface pointers are added here,
	 * be sure and update fd_free() to release them.
	 */
	struct oskit_stream	*stream;
	struct oskit_posixio	*posixio;
	struct oskit_openfile	*openfile;
	struct oskit_socket	*socket;
	struct oskit_ttystream	*ttystream;
	struct oskit_asyncio	*asyncio;
	struct oskit_bufio	*bufio;
	oskit_s32_t		selecting_on;
	struct oskit_listener	*listener;
};
typedef struct fd fd_t;


/* This is the array in which we keep the file descriptors */
extern fd_t *fd_array;
extern int   fd_arraylen;


OSKIT_BEGIN_DECLS


/*** Initialization functions ***/
/*
 * The client must call these functions as appropriate
 * in order to hook the minimal fd layer up
 * to the actual services it exports.
 * The client only needs to initialize the services it actually uses.
 */

/*
 * Set file descriptors 0, 1, and 2 (stdin, stdout, and stderr)
 * to refer to a fake "console stream" that just uses putchar and getchar.
 */
oskit_error_t fd_set_console(void);

/*
 * Interface to switch out the select function when running multi-threaded.
 */
typedef int (*oskit_select_function_t)
	    (int n, fd_set *in,
	     fd_set *out, fd_set *exc, struct timeval *tout);

void fd_set_select_function(oskit_select_function_t);

/*** Internal functions ***/
/*
 * The client normally doesn't need to use these directly,
 * but it can if it wants to extend the layer or do other special things.
 */

/*
 * These functions by default do nothing, but can be overridden by the
 * C library when it detects that there is a registered lock manager,
 * and thus needs to provide proper locking.
 */
void	fd_lock(void);
void	fd_unlock(void);
void	fd_lock_init(void);

/*
 * Allocate a new file descriptor and return the descriptor number.
 * The descriptor number allocated will be at least min_fd
 * (this is to support F_DUPFD; just pass 0 to allocate any descriptor).
 * Returns -1 and errno set correctly if no fd can be allocated.
 * FD layer must be locked throughout.
 */
int	fd_alloc(oskit_iunknown_t *obj, int min_fd);

/*
 * Free a file descriptor by releasing and zeroing all its object references.
 * FD layer must be locked throughout.
 */
int	fd_free(int fd);

/*
 * test whether fd is invalid or not open
 */

#define FD_BAD(fd) (fd < 0 || fd >= fd_arraylen || fd_array[fd].obj == 0)

/*
 * Lock the file descriptor layer,
 * then verify that a file descriptor number is valid and open.
 */
#define FD_CHECK(fd)      						\
	fd_lock();							\
	if (FD_BAD(fd)) {						\
		fd_unlock();						\
		errno = EBADF;						\
		return -1;						\
	}

/*
 * check that the underlying object supports a particular interface, 
 * querying for it if needed
 */
#define FD_HAS_INTERFACE(fd, interface)					\
	(fd_array[fd].interface != 0 ||					\
	    oskit_iunknown_query(fd_array[fd].obj,			\
			&oskit_##interface##_iid,			\
			(void**)&fd_array[fd].interface) == 0) 		\

/*
 * Verify that a file descriptor is valid and open,
 * and that its underlying object supports a particular interface.
 * If the macro completes (as opposed to returning an error),
 * then the appropriate cached interface pointer is valid.
 */
#define FD_CHECK_INTERFACE(fd, interface, error)			\
	FD_CHECK(fd);							\
	if (!FD_HAS_INTERFACE(fd, interface)) {				\
		fd_unlock();						\
		errno = error;						\
		return -1;						\
	}								\
	fd_unlock();
	
#define FD_CHECK_STREAM(fd)	FD_CHECK_INTERFACE(fd, stream, EBADF)
#define FD_CHECK_OPENFILE(fd)	FD_CHECK_INTERFACE(fd, openfile, EBADF)
#define FD_CHECK_SOCKET(fd)	FD_CHECK_INTERFACE(fd, socket, ENOTSOCK)
#define FD_CHECK_TTYSTREAM(fd)	FD_CHECK_INTERFACE(fd, ttystream, ENOTTY)
#define FD_CHECK_ASYNCIO(fd)	FD_CHECK_INTERFACE(fd, asyncio, ENOTTY)

/* 
 * a posixio can be obtained using two different ways: it is either
 * supported by the object itself, or, if the object supports the openfile
 * interface, it may be supported by the corresponding file object
 */
#define FD_CHECK_POSIXIO(fd)						\
	FD_CHECK(fd);							\
	if (!FD_HAS_INTERFACE(fd, posixio)) {				\
		if (!FD_HAS_INTERFACE(fd, openfile) ||			\
		    oskit_openfile_getfile(fd_array[fd].openfile, 	\
		   (struct oskit_posixio *)&fd_array[fd].posixio) != 0)	\
		{							\
			fd_unlock();					\
			errno = EBADF;					\
			return -1;					\
		}							\
	}								\
	fd_unlock();

OSKIT_END_DECLS

#endif /* _OSKIT_C_FD_H_ */
