/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * External definitions for CPU inheritance.
 */
#ifndef	_CPUINHERIT_H_
#define _CPUINHERIT_H_

/*
 * CPU inheritance messages
 */
/*
 * Message types.
 */
typedef enum {
	MSG_SCHED_NOTSETYET = 0,
	MSG_SCHED_BLOCK,		/* Thread is blocking */
	MSG_SCHED_UNBLOCK,		/* Thread is requesting reschedule */
	MSG_SCHED_SETSTATE,		/* Change thread state (priority) */
	MSG_SCHED_EXITED,		/* Thread is exiting */
	MSG_SCHED_USERYIELD,		/* User directed yield request */
	MSG_SCHED_YIELD,		/* System directed yield request */
	MSG_SCHED_PREEMPT,		/* Timer based preemption */
	MSG_SCHED_MAXTYPE,
} schedmsg_types_t;

extern char *msg_sched_typenames[];	/* Type strings for debugging */

/*
 * Message format.
 */
typedef struct schedmsg {
	schedmsg_types_t	type;	/* Message type code */
	pthread_t		tid;	/* Thread under operation */
	oskit_u32_t		opaque;	/* Extra data if appropriate */
} schedmsg_t;

/*
 * Wakeup conditions for pthread_sched_donate().
 */
typedef enum {
	WAKEUP_NOTSETYET  = 0,
	WAKEUP_NEVER,			/* Internal. DO NOT USE! */
	WAKEUP_ON_BLOCK,
	WAKEUP_ON_SWITCH,
	WAKEUP_ALWAYS,
} sched_wakecond_t;

/*
 * Core cpu inheritance prototypes.
 */
/*
 * Wait for a new message to arrive. Called from a scheduler when it has
 * nothing to run and is waiting for a message to tell it to do something.
 *
 * In the current implementation, this call returns periodically so that
 * the caller can test for cancelation. Would a timeout be better? This
 * would of course be inconsistent with the timeout given to donate, since
 * this would be real time, not virtual time. 
 */
void		pthread_sched_wait(void);

/*
 * Non-blocking call to retrieve a new scheduling message.
 */
int		pthread_sched_msg_recv(schedmsg_t *pmsg);

/*
 * Donate CPU to a thread. Called by schedulers to donate CPU time to
 * the next thread it wants to run. Wakeup condition indicates under
 * what circumstance the thread donating should be woken up when the
 * target blocks. Timeout is in milliseconds; control returns if the
 * timeout expires.
 *
 * The return value indicates if the donation was actually made. If
 * the thread being donated to has blocked, then the donation fails
 * and the caller needs to do something appropriate.
 */
int		pthread_sched_donate(pthread_t tid,
			sched_wakecond_t cond, oskit_s32_t timeout);

/*
 * Finalize thread exit. Schedulers call this when a thread exits. Tells
 * the underlying thread system to handle the details of deallocation.
 */
void		pthread_sched_exit_finalize(pthread_t tid);

/*
 * A scheduler starts up and tells the core that its a scheduler, and
 * optionally provides a callback routine to be called each time a new
 * thread is created as a child of that scheduler.
 */
void		pthread_sched_become_scheduler(
			void (*initializer)(pthread_t tid,
					    const pthread_attr_t *attr));

/*
 * Alter the scheduling parameters for a thread. Currently, just one
 * opaque value is supported (tickets, priority, reservation, etc).
 * This will need to beef up at some point.
 */
int		pthread_sched_setstate(pthread_t tid, int opaque);

/*
 * Additional attribute support.
 */
/*
 * Set an opaque value that represents scheduling information.
 * ie: priority, tickets, reservation, etc.
 */
int		pthread_attr_setopaque(pthread_attr_t *attr, oskit_u32_t op);

/*
 * Set the parent scheduler for a newly created thread. By default, a
 * thread will be created with the same parent scheduler as the thread
 * that created it. Set it to a non-zero TID to reparent the new thread.
 */
int		pthread_attr_setscheduler(pthread_attr_t *attr, pthread_t tid);

/*
 * Currently defined schedulers.
 */
/*
 * Fixed Priority. Either FIFO or RR, as in POSIX.
 */
int		create_fixedpri_scheduler(pthread_t *tid,
			const pthread_attr_t *attr, int preemptible);

/*
 * Lottery
 */
int		create_lotto_scheduler(pthread_t *tid,
			const pthread_attr_t *attr, int preemptible);

/*
 * Stride
 */
int		create_stride_scheduler(pthread_t *tid,
			const pthread_attr_t *attr, int preemptible);

/*
 * Rate Monotonic
 */
int		create_ratemono_scheduler(pthread_t *tid,
			const pthread_attr_t *attr, int preemptible);

/*
 * Root scheduler prototype to get it bootstrapped.
 */
void		bootstrap_root_scheduler(pthread_t tid, int preemptible,
			void (**function)(void *), void **argument);

/*
 * Gross
 */
int		pthread_setspecific_bytid(pthread_t tid,
			pthread_key_t key, const void *value);
void	       *pthread_getspecific_bytid(pthread_t tid, pthread_key_t key);


/*
 * Debug stuff.
 */
extern int			cpuinherit_debug;
int				cpuprintf(const char *__format, ...);

#define CPUDEBUG_TRUE		0x8000
#define CPUDEBUG_CORE0		0x0001
#define CPUDEBUG_CORE1		0x0002
#define CPUDEBUG_CORE2		0x0004
#define CPUDEBUG_CORE3		0x0008
#define CPUDEBUG_FIXEDPRI	0x0010
#define CPUDEBUG_RATEMONO	0x0020
#define CPUDEBUG_LOTTO		0x0040
#define CPUDEBUG_STRIDE		0x0080

#define CPUDEBUG(cond, args...) ({		\
	if (cpuinherit_debug & CPUDEBUG_##cond)	\
		cpuprintf(##args);		\
	})

#define CPUDEBUG_ISSET(cond)	(cpuinherit_debug & CPUDEBUG_##cond)

#endif /* _CPUINHERIT_H_ */
