/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#ifdef	CPU_INHERIT
/*
 * Full Blown CPU Inheritance scheduler.
 */
#include <threads/pthread_internal.h>
#include <stdlib.h>

void
pthread_sched_send_message(resched_flags_t reason,
			   pthread_thread_t *pscheduler,
			   pthread_thread_t *pthread);

/*
 * There is a single root scheduler.
 */
pthread_thread_t        *pthread_root_scheduler = NULL_THREADPTR;
int			 cpuinherit_debug       = CPUDEBUG_TRUE;

static inline int
ondonorq(pthread_thread_t *pthread)
{
	return (int) pthread->donors_chain.next;
}

static inline void
donorq_enter(queue_head_t *q, pthread_thread_t *pthread)
{
	queue_enter(q, pthread, pthread_thread_t *, donors_chain);
}

static inline void
donorq_remove(queue_head_t *q, pthread_thread_t *pthread)
{
	queue_remove(q, pthread, pthread_thread_t *, donors_chain);
	pthread->donors_chain.next = 0;
}

/*
 * For debugging.
 */
char *msg_sched_typenames[] = {
	"BADTYPE",
	"BLOCK",
	"UNBLOCK",
	"SETSTATE",
	"EXITED",
	"USERYIELD",
	"YIELD",
	"PREEMPT",
	"MAXTYPE",
};

char *msg_sched_flagnames[] = {
	"NOTSETYET",
	"DONATING",
	"RUNNING",
	"READY",
	"WAITING",
};

#define STATESTR(p)	(msg_sched_flagnames[(p)->schedflags])

/*
 * There are two sets of functions. The first are routines exported to
 * the core scheduling system. The second are routines exported to the
 * schedulers. The first opertate on thread pointers directly. The second
 * indirect through thread TIDs.
 */

/*
 * Initialize the scheduler subsystem.
 */
void
pthread_init_scheduler(void)
{
	char	*bp;

	if ((bp = getenv("CPUDEBUG")) != NULL)
		cpuinherit_debug = strtol(bp, NULL, 0);

	printf("CPUINHERIT_DEBUG = 0x%x\n", cpuinherit_debug);
}

/*
 * A thread is being resumed. Send a message to the threads scheduler.
 */
int 
pthread_sched_setrunnable(pthread_thread_t *pthread)
{
	pthread_sched_wakeup(pthread);

	/*
	 * Return 1 to force a pthread_yield call.
	 */
	return 1;
}

/*
 * The current thread is exiting. Send a message to its scheduler,
 * and then switch back into the scheduler.
 */
void
pthread_sched_exit(void)
{
	pthread_thread_t	*pthread = CURPTHREAD();

	/*
	 * Special case is the root scheduler. When it exits, just reset.
	 */
	if (pthread == pthread_root_scheduler)
		exit(0);

	schedmsg_enqueue(pthread->scheduler->schedqueue,
			 MSG_SCHED_EXITED, pthread->tid, 0);

	pthread_sched_reschedule(RESCHED_INTERNAL, 0);

	/*
	 * Never returns.
	 */
}

/*
 * Change the state of a thread. Send a message to the threads scheduler.
 * When pthread_yield is called, another message is going to be generated,
 * buts thats okay. This first message will change the thread state and
 * the second message will be the yield message for whatever thread is
 * actually running (might be different than the thread being changed).
 */
int
pthread_sched_change_state(pthread_thread_t *pthread, int newprio, int policy)
{
	/*
	 * XXX: Cannot change the policy after startup.
	 */
	schedmsg_enqueue(pthread->scheduler->schedqueue,
			 MSG_SCHED_SETSTATE, pthread->tid, newprio);

	pthread_sched_wakeup(pthread->scheduler);
	
	/*
	 * Return 1 to force a pthread_yield call.
	 */
	return 1;
}

/*
 * Helper function for creating schedulers. Force a switch into a specific
 * thread. Nothing special about it.
 */
void
pthread_sched_switchto(pthread_thread_t *pthread)
{
	int	s;
	
	s = splhigh();
	pthread_lock(&pthread->lock);
	thread_switch(pthread, &pthread->lock, &CURPTHREAD());
	splx(s);
}

/*
 * Switch out of an application thread, back into its scheduler. Return
 * value to indicate whether a switch actually happened.
 */
int
pthread_sched_reschedule(resched_flags_t reason, pthread_lock_t *plock)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	pthread_thread_t	*inheritor, *donator;
	int			s, mtype;

	s = splhigh();

	/*
	 * Need to lock the thread lock and then release the provided
	 * spinlock. This provides atomicity with regards to the thread
	 * that is switching out since in all cases the lock is either
	 * the thread lock, or a queue that it just put itself on.
	 */
	if (plock != &(pthread->lock)) {
		pthread_lock(&(pthread->lock));
		
		if (plock)
			pthread_unlock(plock);
	}

	/*
	 * See what donator should receive the message and get restarted.
	 * Go back up the inheritance chain. Stop at the root scheduler
	 * (which will have set WAKEUP_ALWAYS) or when some intermediate
	 * thread says it wants to be woken up.
	 */
	donator   = pthread->inherits_from;
	inheritor = pthread;
	assert(inheritor != donator);

	/*
	 * Better be a donator unless its the root scheduler.
	 */
	assert(donator || (pthread == pthread_root_scheduler));

	/*
	 * The current thread state is set accordingly.
	 */
	if (reason == RESCHED_BLOCK)
		pthread->schedflags = SCHED_WAITING;
	else
		pthread->schedflags = SCHED_READY;

	/*
	 * Go back through the inheritance links, stopping at the first
	 * thread to indicate that it wants to be woken up.
	 */
	while (donator) {
		pthread_unlock(&inheritor->lock);
		pthread_lock(&donator->lock);

		if (donator->wakeup_cond == WAKEUP_ALWAYS ||
		    donator->wakeup_cond == WAKEUP_ON_BLOCK)
			break;

		/*
		 * With each step back through the active chain, set
		 * the thread state accordingly. Blocks are different
		 * then yields in that another thread should be able
		 * to donate to a thread that yielded. However, blocked
		 * chains should cause a wakeup for any thread along
		 * the chain.
		 */
		if (reason == RESCHED_BLOCK)
			donator->schedflags = SCHED_WAITING;
		else
			donator->schedflags = SCHED_READY;

		inheritor = donator;
		donator   = donator->inherits_from;
	}
	
	if (donator == NULL_THREADPTR) {
		/*
		 * Special case is the root scheduler. If there is no
		 * thread to run, and not even the root scheduler has
		 * something to do, spin waiting for messages to arrive
		 * on the root schedulers message queue.
		 *
		 * In other words, this is the idle loop.
		 */
		pthread_root_scheduler->schedflags = SCHED_WAITING;
		pthread_unlock(&pthread_root_scheduler->lock);
		spllow();
		
		while (pthread_root_scheduler->schedflags != SCHED_READY)
			pthread_delay();

		donator = pthread_root_scheduler;
		splhigh();
		pthread_lock(&donator->lock);
		goto dispatch;
	}

	/*
	 * Send a message to the scheduler. These messages do not encode a
	 * thread ID, since as far as the scheduler is concerned, control
	 * has returned from the thread it donated to. This is the case
	 * even if it was some thread further down the chain that asked for
	 * a reschedule; the scheduler only sees the state of its donation
	 * changed.
	 *
	 * The only reason to distingiush between all these reschedule
	 * reasons in the message is so that we can implement a POSIX style
	 * scheduler. This is probably unnecessary, and makes the
	 * schedulers more compicated then they need to be. Note that the
	 * PREEMPT message is sent from the clocktick code, not here!
	 */
	switch (reason) {
	case RESCHED_USERYIELD:
		mtype = MSG_SCHED_USERYIELD;
		break;
	case RESCHED_YIELD:
		mtype = MSG_SCHED_YIELD;
		break;
	case RESCHED_BLOCK:
		mtype = MSG_SCHED_BLOCK;
		break;
	default:
		mtype = 0;
	}

	if (mtype) 
		schedmsg_enqueue(donator->schedqueue, mtype, 0, 0);

dispatch:
	reason = 0;
	assert(donator);

	thread_switch(donator, &donator->lock, &CURPTHREAD());

	assert(pthread->inherits_from || pthread == pthread_root_scheduler);
	
	/*
	 * Thread has switched back in. Time to look for exceptions before
	 * letting it return to whatever it was doing before it yielded.
	 */
	pthread_lock(&pthread->lock);
	pthread->schedflags = SCHED_RUNNING;
	
	if (pthread->sleeprec) {
		/*
		 * Thread is returning to an osenv_sleep. The thread must
		 * be allowed to return, even if it was killed, so the driver
		 * state can be cleaned up. The death will be noticed later.
		 */
		;
	}
	else if ((pthread->flags &
		  (THREAD_KILLED|THREAD_EXITING)) == THREAD_KILLED) {
		/*
		 * Thread was canceled. Time to actually reap it, but only
		 * if its not already in the process of exiting.
		 *
		 * XXX: The problem is if the process lock is still
		 * held.  Since the rest of the oskit is not set up to
		 * handle cancelation of I/O operations, let threads
		 * continue to run until the process lock is released.
		 * At that point the death will be noticed.
		 */
		if (! osenv_process_locked()) {
			pthread_unlock(&pthread->lock);
			pthread_reap();

			/*
			 * Never returns
			 */
		}
	}

	pthread_unlock(&pthread->lock);
	splx(s);
	return 1;
}

/*
 * Wakeup call.
 */
void
pthread_sched_wakeup(pthread_thread_t *pthread)
{
	int			s;
	pthread_thread_t	*pscheduler = pthread->scheduler;
	pthread_thread_t	*inheritor, *ptmp;

	s = splhigh();
	pthread_lock(&pthread->lock);
	
	CPUDEBUG(CORE0,
		 "pthread_sched_wakeup: "
		 "C:0x%x(%d)/%s T:0x%x(%d)/%s S:0x%x(%d)/%s\n",
		 (int) CURPTHREAD(), CURPTHREAD()->tid, STATESTR(CURPTHREAD()),
		 (int) pthread, pthread->tid, STATESTR(pthread),
		 (int) pscheduler,
		 (pscheduler ? pscheduler->tid : 0),
		 (pscheduler ? STATESTR(pscheduler) : ""));


	/*
	 * Thread is currently ready to run. The wrinkle is if the thread
	 * was part of a donation chain that yielded, or a donation chain
	 * that was interrupted by a wakeup. There will be a donating_to
	 * link, which must be severed so that when the chain is run again,
	 * the thread in question will run, rather then following the
	 * chain.
	 */
	if (pthread->schedflags == SCHED_READY) {
		if (pthread->donating_to) {
			CPUDEBUG(CORE3,
				 "PSW0: 0x%x(%d) 0x%x 0x%x\n",
				 (int) pthread, pthread->tid,
				 (int) pthread->donating_to,
				 (int) pthread->scheduler);
			
			pthread->donating_to = NULL_THREADPTR;
			pthread->wakeup_cond = WAKEUP_ALWAYS;
		}
		pthread_unlock(&pthread->lock);
		splx(s);
		return;
	}

	/*
	 * Thread is donating, which means its on the active chain.
	 */
	if (pthread->schedflags == SCHED_DONATING) {
		pthread->schedflags = SCHED_READY;
		
		/*
		 * Follow the chain down, setting each one along the way
		 * to the ready state instead of donating. 
		 */
		inheritor = pthread->donating_to;
		pthread_lock(&inheritor->lock);
		
		while (inheritor->donating_to) {
			inheritor->schedflags = SCHED_READY;

			ptmp      = inheritor;
			inheritor = inheritor->donating_to;
			pthread_unlock(&ptmp->lock);
			pthread_lock(&inheritor->lock);
		}
		
		/*
		 * Got to the end of the chain. Severe the chain at the
		 * thread being woken up. Point the last thread in the
		 * chain (usually the thread actually running) back at the
		 * this thread. This means that when the running thread
		 * switches out, it will follow its link back to this
		 * thread.
		 */
		pthread->donating_to     = NULL_THREADPTR;
		pthread->wakeup_cond     = WAKEUP_ALWAYS;
		inheritor->inherits_from = pthread;

		/*
		 * The wrinkle is if the last thread in the chain is not
		 * the currently running thread. That means the chain
		 * below this point was already woken up (it was severed).
		 * Need to entend the backwards chain so that when the
		 * currently running thread switches out, it follows the
		 * links back to this thread (along multiple hops). 
		 */
		if (inheritor != CURPTHREAD())
			inheritor->wakeup_cond = WAKEUP_NEVER;

		CPUDEBUG(CORE3,
			 "PSW1: 0x%x(%d)\n", (int) inheritor, inheritor->tid);
			
		pthread_unlock(&pthread->lock);
		pthread_unlock(&inheritor->lock);
		splx(s);
		return;
	}

	/*
	 * The thread is not on the active chain, and its not in the READY
	 * state (already woken up). Need to follow the chain back until we
	 * get to a thread *is* DONATING.
	 */
	CPUDEBUG(CORE1, "PSW2:\n");

	/*
	 * Want to wake up either the thread donating to it, or its scheduler.
	 * Of course, they might be the same.
	 */
	if (pthread->inherits_from &&
	    pthread->inherits_from != pscheduler) {
		/*
		 * A non-scheduler thread was donating to another thread,
		 * and the inheritor blocked. Now the inheritor is being
		 * woken up. Need to wakeup the donator in the hopes that
		 * it can provide some more CPU.
		 */
		printf("PSW3: 0x%x 0x%x 0x%x\n",
		       (int) pthread, (int) pscheduler,
		       (int) pthread->inherits_from);
		
		pthread->schedflags = SCHED_READY;
		inheritor = pthread->inherits_from;
		pthread_unlock(&pthread->lock);
		
		pthread_sched_wakeup(inheritor);
	}
	else if (pscheduler) {
		pthread->schedflags = SCHED_READY;
		pthread_unlock(&pthread->lock);
		pthread_lock(&pscheduler->lock);
		
		/*
		 * If the thread waking up is different than the thread
		 * the scheduler is currently donating to, this implies
		 * two (or more) threads, and thus a conflict. Set the
		 * wakeup_cond back to WAKEUP_ALWAYS so that the scheduler
		 * will be woken up.
		 */
		if (pscheduler->donating_to &&
		    pscheduler->donating_to != pthread) {
			pscheduler->wakeup_cond = WAKEUP_ALWAYS;

			CPUDEBUG(CORE2,
				 "PSW4: 0x%x\n",
				 (int) pscheduler->donating_to);
		}

		/*
		 * If the scheduler is not going to be woken up, then
		 * don't send it a message.
		 */
		if (pscheduler->wakeup_cond == WAKEUP_ALWAYS ||
		    pscheduler->wakeup_cond == WAKEUP_ON_BLOCK) {
			CPUDEBUG(CORE2,
				 "PSW5: 0x%x(%d) 0x%x(%d)\n",
				 (int) pscheduler, pscheduler->tid,
				 (int) pthread, pthread->tid);

			schedmsg_enqueue(pscheduler->schedqueue,
					 MSG_SCHED_UNBLOCK, pthread->tid, 0);
		}
		pthread_unlock(&pscheduler->lock);
		pthread_sched_wakeup(pscheduler);
	}
	else {
		/*
		 * Hit the root scheduler. 
		 */
		pthread_root_scheduler->schedflags = SCHED_READY;
		pthread_unlock(&pthread->lock);
	}
	splx(s);
}

/*
 * Priority inheritance entrypoint. Donate CPU time to a thread being
 * waited for. Called from the mutex code.
 */
void
pthread_sched_waiton(pthread_thread_t *waiting_on,
		     queue_head_t *queue, pthread_lock_t *plock)
{
	pthread_thread_t	*pthread = CURPTHREAD();

	queue_enter(queue, pthread, pthread_thread_t *, chain);
	pthread_unlock(plock);	/* XXX SMP problem */

	printf("pthread_sched_waiton: h:0x%x(%d) p:0x%x(%d)\n",
	       (int) waiting_on, waiting_on->tid,
	       (int) pthread, pthread->tid);
	
	if (!pthread_sched_donate(waiting_on->tid, WAKEUP_NEVER, 0))
		pthread_sched_reschedule(RESCHED_BLOCK, 0);
}

/*
 * Clock tick handler. Called from pthread_interrupt (interrupts disabled).
 */
void
pthread_sched_clocktick(void)
{
	pthread_thread_t	*pthread;
	int			needsoftint = 0;

	if (PREEMPTBLOCKED)
		return;

	pthread = CURPTHREAD();

	/*
	 * Look for preemption along the active chain. These are threads
	 * that donated with a timeout value. Decrement the timeout and
	 * schedule any threads that expire. 
	 */
	while (pthread) {
		if (pthread->timeout) {
			pthread->timeout -= PTHREAD_TICK;
			
			if (pthread->timeout <= 0) {
				pthread->timeout = 0;
				
				pthread->wakeup_cond = WAKEUP_ALWAYS;
				schedmsg_enqueue(pthread->schedqueue,
						 MSG_SCHED_PREEMPT, 0, 0);

				pthread_sched_wakeup(pthread);

				needsoftint = 1;
			}
		}
		pthread = pthread->inherits_from;
	}

	if (needsoftint && CURPTHREAD() != pthread_root_scheduler)
		softint_request(SOFTINT_TIMEOUT);
}

/*
 * These are the routines that are exported to schedulers.
 */
/*
 * A scheduler tells us its a scheduler, and optionally provides a callback
 * routine to call on each new thread created for that scheduler. That
 * provides a hook that allows the scheduler to initialize its internal state
 * before it sees that thread for the first time in the unblock call that
 * is used to get it started. 
 */
void
pthread_sched_become_scheduler(
	void (*initializer)(pthread_t tid, const pthread_attr_t *attr))

{
	pthread_thread_t	*pthread = CURPTHREAD();
	
	/*
	 * Number the hierarchy.
	 */
	if (pthread->scheduler)
		pthread->schedlevel = pthread->scheduler->schedlevel + 1;

	pthread->schedinit = initializer;
}

/*
 * Thread has completed exiting. Time to reap it or awaken the threads
 * waiting to join with it.
 */
void
pthread_sched_exit_finalize(pthread_t tid)
{
	pthread_thread_t	*pthread;
	int			s, resched = 0;

	if (tid < 0 || tid >= THREADS_MAX_THREAD ||
	    ((pthread = tidtothread(tid)) == NULL_THREADPTR))
		panic("pthread_sched_exit_finalize: Bad TID: %d\n", tid);

	CPUDEBUG(CORE0,
		 "pthread_sched_exit_finalize(%d): Thread 0x%x exited\n",
		 pthread_self(), (int) pthread);

	s = splhigh();
	pthread_lock(&pthread->lock);

	if (pthread->flags & THREAD_DETACHED) {
		splx(s);
		pthread_destroy_internal(pthread);
	}
	else {
		/*
		 * Schedule all the threads waiting.
		 */
		if (! queue_empty(&pthread->joinq)) {
			pthread_resume_allQ(&pthread->joinq);

			if (CURPTHREAD() != pthread_root_scheduler)
				resched = 1;
		}
		pthread->flags |= THREAD_EXITED;
		pthread_unlock(&pthread->lock);
		splx(s);
	}

	if (resched)
		pthread_yield();		
}

/*
 * Change the scheduling parameters for a thread. Currently a hack that
 * sends a message to the threads scheduler containing a single opaque
 * value that hopefully makes sense to the scheduler. This will need to
 * beef up at some point.
 */
int
pthread_sched_setstate(pthread_t tid, int opaque)
{
	pthread_thread_t  *pthread;

	if (tid < 0 || tid >= THREADS_MAX_THREAD ||
	    ((pthread = tidtothread(tid)) == NULL_THREADPTR))
		return EINVAL;

	schedmsg_enqueue(pthread->scheduler->schedqueue,
			 MSG_SCHED_SETSTATE, pthread->tid, opaque);

	pthread_sched_wakeup(pthread->scheduler);

	return 0;
}

/*
 * Look for a scheduling message. Called by schedulers.
 */
int
pthread_sched_msg_recv(schedmsg_t *pmsg)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	schedmsg_queue_t	*pqueue  = pthread->schedqueue;

	if (schedmsg_dequeue(pqueue, pmsg))
		return 1;

	return 0;
}

/*
 * This is the basic scheduling call invoked to donate CPU time to a thread.
 * The return value indicates whether the donation actually occurred (thread
 * was ready to run). This tells the caller if it needs to just block.
 */
int
pthread_sched_donate(pthread_t tid, sched_wakecond_t wakeup_cond,
		     oskit_s32_t timeout)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	pthread_thread_t	*pnext, *pchild;
	int			s;

	if (tid < 0 || tid >= THREADS_MAX_THREAD ||
	    ((pchild = tidtothread(tid)) == NULL_THREADPTR))
		panic("pthread_sched_donate: Bad TID: %d\n", tid);

	s = splhigh();
	pthread_lock(&pthread->lock);
	pnext = pchild;

	CPUDEBUG(CORE3,
		 "PSD0: 0x%x(%d) 0x%x(%d) %s\n",
		 (int) pthread, pthread->tid,
		 (int) pchild, pchild->tid, STATESTR(pchild));

	/*
	 * If the thread is not ready to accept this new donation, add
	 * it to donors list; it might be used later.
	 */
	if (pchild->schedflags != SCHED_READY) {
		CPUDEBUG(CORE3,
			 "PSD1: 0x%x(%d) 0x%x(%d) %s\n",
			 (int) pthread, pthread->tid,
			 (int) pchild, pchild->tid, STATESTR(pchild));

		donorq_enter(&pchild->donors, pthread);

		pthread_unlock(&pthread->lock);
		splx(s);
		return 0;
	}

	/*
	 * Is another thread already donating to this thread? Make that
	 * thread an additional donor. It might be used later if this
	 * new donator runs out of CPU.
	 */
	if (pchild->inherits_from) {
		CPUDEBUG(CORE3,
			 "PSD2: 0x%x(%d) 0x%x(%d) 0x%x(%d)\n",
			 (int) pthread, pthread->tid,
			 (int) pchild, pchild->tid,
			 (int) pchild->inherits_from,
			 pchild->inherits_from->tid);

		donorq_enter(&pchild->donors, pchild->inherits_from);
	}

	/*
	 * If the thread being switched into is part of an old active
	 * chain, decend the chain setting each thread to DONATED.
	 * Set the back pointer too since it might no longer be valid
	 * if another thread donated to the chain at some point.
	 * Stop when we hit a thread with WAKEUP_ALWAYS set, or at the end.
	 */
	if (pnext->donating_to) {
		while (pnext->donating_to) {
			if (pnext->schedflags != SCHED_READY)
				panic("pthread_sched_donate SCHED_READY");

			if (pnext->wakeup_cond == WAKEUP_ALWAYS ||
			    pnext->wakeup_cond == WAKEUP_ON_BLOCK)
				break;
		
			pnext->schedflags = SCHED_DONATING;
			pnext->donating_to->inherits_from = pnext;
			
			CPUDEBUG(CORE3,
				 "PSD3: "
				 "Followed chain from 0x%x(%d) to 0x%x(%d)\n",
				 (int) pnext, pnext->tid,
				 (int) pnext->donating_to,
				 pnext->donating_to->tid);

			pnext = pnext->donating_to;
		}
	}

	/*
	 * XXX: Make sure this donator is not in the donors queue.
	 */
	if (ondonorq(pthread)) {
		donorq_remove(&pchild->donors, pthread);

		CPUDEBUG(CORE3,
			 "PSD4: 0x%x(%d) 0x%x(%d)\n",
			 (int) pthread, pthread->tid,
			 (int) pchild, pchild->tid);
	}

	/*
	 * Set up the links.
	 */
	pthread->schedflags   = SCHED_DONATING;
	pthread->donating_to  = pchild;
	pthread->wakeup_cond  = wakeup_cond;
	pthread->timeout      = timeout;
	pchild->inherits_from = pthread;
	pnext->schedflags     = SCHED_RUNNING;

	thread_switch(pnext, &pthread->lock, &CURPTHREAD());

	/*
	 * Undo the donating link. Leave the back link alone since another
	 * thread could have donated to the child if the chain was preempted.
	 */
	pthread->timeout      = 0;
	pthread->donating_to  = NULL_THREADPTR;
	if (pchild->inherits_from == pthread)
		pchild->inherits_from = NULL_THREADPTR;

	/*
	 * XXX: Make sure this donator is not in the donors queue.
	 */
	if (ondonorq(pthread)) {
		donorq_remove(&pchild->donors, pthread);

		CPUDEBUG(CORE3,
			 "PSD6: 0x%x(%d) 0x%x(%d)\n",
			 (int) pthread, pthread->tid,
			 (int) pchild, pchild->tid);
	}

	pthread->schedflags = SCHED_RUNNING;
	splx(s);

	return 1;
}

/*
 * A scheduler thread has nothing to do. It will block waiting for a
 * message to arrive with something for it to do. Pass control back
 * up the line, telling its scheduler that it has blocked. 
 */
void
pthread_sched_wait(void)
{
	pthread_thread_t	*pthread = CURPTHREAD();
	schedmsg_queue_t	*pqueue  = pthread->schedqueue;
	int			s;

	s = splhigh();

	/*
	 * First look for a message before blocking.
	 */
	if (! schedmsg_queue_empty(pqueue)) {
		splx(s);
		return;
	}
	pthread_lock(&pthread->lock);

	/*
	 * Must wakeup when something interesting happens.
	 */
	pthread->wakeup_cond = WAKEUP_ALWAYS;

	/*
	 * Special thread state for cancelation code.
	 */
	pthread->flags |= THREAD_MSGWAIT;

	pthread_sched_reschedule(RESCHED_BLOCK, &pthread->lock);

	pthread_lock(&pthread->lock);
	pthread->flags &= ~THREAD_MSGWAIT;
	pthread_unlock(&pthread->lock);
	
	splx(s);
}

/*
 * Debugging!
 */
#include <stdarg.h>
#include <unistd.h>

int
cpuprintf(const char *fmt, ...)
{
	va_list	args;
	int err, s;
	char buf[256];

	s = splhigh();
	va_start(args, fmt);
	err = vsnprintf(buf, 256, fmt, args);
	va_end(args);
	write(1, buf, strlen(buf));
	splx(s);

	return err;
}

#endif /* CPU_INHERIT */
