/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Internal definitions for CPU inheritance.
 */
#ifndef	_PTHREAD_CPUINHERIT_H_
#define _PTHREAD_CPUINHERIT_H_

#include <oskit/threads/cpuinherit.h>

/*
 * Scheduler flags specific to CPU inheritance. See pthread_sched_reschedule.
 */
typedef enum {
	SCHED_NOTSETYET    = 0,
	SCHED_DONATING,		/* Dontaing CPU to another thread */
	SCHED_RUNNING,		/* Thread has the CPU */
	SCHED_READY,		/* Thread is ready for the CPU */
	SCHED_WAITING,		/* Thread is waiting for something */
} schedflags_t;

/*
 * A message queue. This is a statically sized circular queue. This may 
 * not be ideal, but I cannot stomach calling the allocator and deallocator.
 */
typedef struct schedmsg_queue {
	pthread_lock_t	lock;
	oskit_u32_t	head;
	oskit_u32_t	tail;
	oskit_u32_t	mask;
	int		size;
	schedmsg_t	entries[1];
} schedmsg_queue_t;

/*
 * There is a single root scheduler.
 */
extern struct pthread_thread	*pthread_root_scheduler;

/*
 * Prototypes for the message functions.
 */
/*
 * Allocate and deallocated message queues. Queues are currently fixed
 * size circular buffers of message entries. Data is copied in and out.
 */
schedmsg_queue_t *schedmsg_queue_allocate(int size);
void		schedmsg_queue_deallocate(schedmsg_queue_t *queue);

/*
 * Test if a message queue has any pending messages.
 */
int		schedmsg_queue_empty(schedmsg_queue_t *queue);

/*
 * Enqueue a message on a message queue. The data values are the TID of
 * the thread being operated on, and an optional opaque value that is
 * generally message specific. 
 */
void		schedmsg_enqueue(schedmsg_queue_t *queue,
			schedmsg_types_t type,
			pthread_t tid, oskit_u32_t opaque);

/*
 * Dequeue are return the next available message in a message queue.
 * Since messages do not have their own storage, a message buffer is
 * provided to copyout the message contents.
 */
int		schedmsg_dequeue(schedmsg_queue_t *queue, schedmsg_t *msg);

/*
 * Prototypes for main scheduler functions. These are internal to the thread
 * system. The external calls are defined in oskit/threads/cpuinherit.h
 */
void		pthread_sched_wakeup(struct pthread_thread *pthread);
void		pthread_sched_exit(void);
void		pthread_sched_switchto(struct pthread_thread *pthread);
void		pthread_sched_waiton(struct pthread_thread *waiting_on,
			queue_head_t *queue, pthread_lock_t *plock);
void		pthread_sched_clocktick(void);

extern char *msg_sched_typenames[];	/* Type strings for debugging */
#endif /* _PTHREAD_CPUINHERIT_H_ */


