/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#ifdef	CPU_INHERIT
/*
 * Full Blown CPU Inheritance scheduler.
 */
#include <threads/pthread_internal.h>
#include <oskit/c/malloc.h>

/*
 * Allocate and initialize a scheduler message queue.
 */
schedmsg_queue_t *
schedmsg_queue_allocate(int size)
{
	int		  logsize;
	schedmsg_queue_t *queue;

	/*
	 * Must be a power of two, so round to next power of two greater
	 * than the given size.
	 */
	for (logsize = 1; logsize < size; logsize *= 2)
		;

	if ((queue = (schedmsg_queue_t *)
	     calloc(1, sizeof(schedmsg_queue_t) +
		    ((logsize - 1) * sizeof(schedmsg_t)))) == NULL)
		panic("schedmsg_queue_allocate: No more memory");

	queue->mask = logsize - 1;
	queue->size = logsize;
	pthread_lock_init(&queue->lock);

	return queue;
}

/*
 * Deallocate a scheduler message queue.
 */
void
schedmsg_queue_deallocate(schedmsg_queue_t *queue)
{
	free(queue);
}

/*
 * Check queue status.
 */
int
schedmsg_queue_empty(schedmsg_queue_t *queue)
{
	int	s, empty = 0;

	s = splhigh();	
	pthread_lock(&queue->lock);

	if (queue->head == queue->tail)
		empty = 1;

	pthread_unlock(&queue->lock);
	splx(s);

	return empty;
}

/*
 * Enqueue a message for delivery.
 */
void
schedmsg_enqueue(schedmsg_queue_t *queue, schedmsg_types_t type,
		 pthread_t tid, oskit_u32_t opaque)
{
	int	s, x;

	s = splhigh();	
	pthread_lock(&queue->lock);

	/*
	 * Queue full?
	 */
	if ((queue->head - queue->tail) == queue->size) {
		pthread_unlock(&queue->lock);
		splx(s);
		panic("schedmsg_enqueue: Queue is full: 0x%x\n", (int) queue);
	}
	
	x = (queue->head++ & queue->mask);
	queue->entries[x].type    = type;
	queue->entries[x].tid     = tid;
	queue->entries[x].opaque  = opaque;

	pthread_unlock(&queue->lock);
	splx(s);
}

/*
 * Return the next available message. The values are placed in the provided
 * message buffer since the actual queue entry cannot be used.
 *
 * Return 0 if none available, 1 otherwise.
 */
int
schedmsg_dequeue(schedmsg_queue_t *queue, schedmsg_t *msg)
{
	int	s, x;

	s = splhigh();	
	pthread_lock(&queue->lock);

	/*
	 * Queue empty?
	 */
	if (queue->head == queue->tail) {
		pthread_unlock(&queue->lock);
		splx(s);
		return 0;
	}
	
	x = (queue->tail++ & queue->mask);
	msg->type    = queue->entries[x].type;
	msg->tid     = queue->entries[x].tid;
	msg->opaque  = queue->entries[x].opaque;

	pthread_unlock(&queue->lock);
	splx(s);
	
	return 1;
}
#endif /* CPU_INHERIT */
