/*
 * Copyright (c) 1996, 1998 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

/*
 * Internal routines for threads package.
 */
#include <threads/pthread_internal.h>

/*
 * This is first routine called for every thread.
 */
void
pthread_start_thread(void (*function)(void *), void *argument)
{
	/*
	 * All threads start out with ints blocked, which must be reset.
	 * The idle thread is run preemption disabled. All other threads
	 * run with both enabled.
	 */
	if (CURPTHREAD() == IDLETHREAD)
		splx(SPLPREEMPT);
	else
		splx(0);

	if (threads_debug)
		printf("pthread_start(%d): 0x%x(%d) f:0x%x a:0x%x\n",
		       THISCPU, (int) CURPTHREAD(), CURPTHREAD()->tid,
		       (int) function, (int) argument);
	
	function(argument);

        /* exit with no error */
        pthread_exit(0);

        /* NOT REACHED */
}

/*
 * The idle thread exists to spin looking for threads to run, and provide
 * a place for pthread_delay() to get called so that in usermode CPU time
 * is not wildly consumed.
 */
void
pthread_idle_function(void *arg)
{
	if (threads_debug)
		printf("pthread_idle(%d): starting\n", THISCPU);

	while (1) {
		if (! machine_intr_enabled())
			panic("pthread idleloop: Interrupts disabled!");
		
		/*
		 * If nothing to schedule call the delay function
		 */
		if (!pthread_sched_reschedule(RESCHED_INTERNAL, 0))
			pthread_delay();
	}
}

/*
 * The reaper thread exists to clean up dead threads.
 */
void
pthread_reaper_function(void *arg)
{
	pthread_thread_t	*pthread;
	int			s;

	if (threads_debug)
		printf("pthread_reaper(%d): starting\n", THISCPU);

	pthread_mutex_lock(&threads_reaperq_mutex);

	while (1) {
		if (! machine_intr_enabled())
			panic("pthread reaperloop: Interrupts disabled!");

		while (queue_empty(&threads_reaperq))
			pthread_cond_wait(&threads_reaperq_cond,
					  &threads_reaperq_mutex);

		queue_remove_first(&threads_reaperq,
				   pthread, pthread_thread_t *, chain);

		pthread_mutex_unlock(&threads_reaperq_mutex);

		if (threads_debug)
			printf("Exited! 0x%x(%d)\n",
			       (int) pthread, pthread->tid);

		s = splhigh();
		pthread_lock(&pthread->lock);

		if (pthread->flags & THREAD_DETACHED) {
			splx(s);
			pthread_destroy_internal(pthread);
		}
		else {
			/*
			 * Schedule all the threads waiting.
			 */
			if (! queue_empty(&pthread->joinq))
				pthread_resume_allQ(&pthread->joinq);

			pthread->flags |= THREAD_EXITED;
			pthread_unlock(&pthread->lock);
			splx(s);
		}

		pthread_mutex_lock(&threads_reaperq_mutex);
	}
}

/*
 * Unblock all the threads in the given queue.
 *
 * The caller should have locked the Q.
 */
void
pthread_resume_allQ(queue_head_t *queue)
{
	pthread_thread_t	*pnext;

	queue_iterate(queue, pnext, pthread_thread_t *, chain) {
		if (pnext == CURPTHREAD())
			panic("pthread_resume_allQ");
		
		pthread_sched_setrunnable(pnext);
	}

	queue_init(queue);
}

/*
 * Remove the highest priority item from the Q and return it.
 *
 * The caller should have locked the Q.
 */
pthread_thread_t *
pthread_dequeue_fromQ(queue_head_t *queue)
{
	pthread_thread_t	*pnext = 0, *ptmp;
	int			maxpri = -1;

	queue_iterate(queue, ptmp, pthread_thread_t *, chain) {
		if (ptmp == CURPTHREAD())
			panic("pthread_resume_fromQ");
		
		if (ptmp->priority > maxpri) {
			maxpri = ptmp->priority;
			pnext  = ptmp;
		}
	}
	queue_remove(queue, pnext, pthread_thread_t *, chain);

	return pnext;
}

/*
 * Remove a specific thread from a Q, but only if that thread is really
 * on the queue.
 *
 * The caller should have locked the Q.
 */
pthread_thread_t *
pthread_remove_fromQ(queue_head_t *queue, pthread_thread_t *pthread)
{
	pthread_thread_t	*ptmp = NULL_THREADPTR;

	queue_iterate(queue, ptmp, pthread_thread_t *, chain) {
		if (ptmp == CURPTHREAD())
			panic("pthread_remove_fromQ");
		
		if (ptmp == pthread) {
			queue_remove(queue, ptmp, pthread_thread_t *, chain);
			return ptmp;
		}
	}

	return NULL_THREADPTR;
}

void
pthread_spin_lock(spin_lock_t *lock)
{
	int i;

	for (i = 0; i < 2000000; i++)
		if (spin_try_lock(lock))
			return;

	panic("spin_lock: "
	      "Lock 0x%x locked for a long time. Maybe you deadlocked!",
	      (int) lock);
}
